#include <ctype.h>
#include <netdb.h>
#include <pthread.h>
#include "mep.h"

char host[NI_MAXHOST];
FILE *logfile = NULL;
pthread_key_t key;

void logging_on(void)
{
  if ((logfile = fopen("sessionlog", "w")) == NULL) {
    fprintf(stderr, "Unable to open log file\n");
    exit(EXIT_FAILURE);
  }
}

void logging_off(void)
{
  if (logfile != NULL) {
    fclose(logfile);
    logfile = NULL;
  }
}

void log_in(char *string)
{
  fprintf(logfile, "=> %s", string);
}

void log_out(char *string)
{
  fprintf(logfile, "<= %s", string);
}

void terminate(void)
{
  pthread_key_delete(key);
  logging_off();
  unlink(UNIX_SOCKET);

  exit(EXIT_SUCCESS);
}

void sendout(char *string)
{
  send(((tsd_t *) pthread_getspecific(key))->desc,
       string, strlen(string), 0);
  log_out(string);
}

/* We accept commands of any length for now... */
char *getstring(int desc, const int crlfonly)
{
  char *string = NULL, *temp;
  int read = -1, size = 0;

  while (read != 0) {
    string = (char *) realloc(string, (size + 21) * sizeof(char));
    temp = string + size;
    size += (read = recv(desc, temp, 20, 0)) > 0 ? read : 0;
    temp[read] = '\0';

    if ((crlfonly == 0 && read == 0 && size > 0) ||
	strncmp(string + size - 2, "\r\n", 2) == 0)
      return string;
  }

  return NULL;
}

int mep_init(char *string)
{
  if (strncasecmp(string, "MEP/0.9", 7) == 0) {
    sendout("OK\r\n");

    return 0;
  } else {
    sendout("ERRO\r\n");

    close(((tsd_t *) pthread_getspecific(key))->desc);
    return 1;
  }
}

void mep_notify(char *string)
{
  sendout("OK\r\n");
}

void netsocket(int desc)
{
  char *string;
  tsd_t *tsd;

  string = (char *) malloc((20 + strlen(host)) * sizeof(char));
  sprintf(string, "OK MEP/0.9 %s ready\r\n", host);
  send(desc, string, strlen(string), 0);
  free(string);

  tsd = (tsd_t *) malloc(sizeof(tsd_t));
  pthread_setspecific(key, tsd);
  tsd->desc = desc;

  while ((string = getstring(desc, 1)) != NULL) {
    log_in(string);

    switch (tolower(string[0])) {
    case 'i':
      if (strncasecmp(string, "INIT", 4) == 0) {
	if (mep_init(strip(string, ' ')) != 0)
	  return;
	break;
      }
    case 'n':
      if (strncasecmp(string, "NOTI", 4) == 0) {
	mep_notify(strip(string, ' '));
	break;
      }
    case 'q':
      if (strncasecmp(string, "QUIT", 4) == 0) {
	mep_quit(strip(string, ' '));
	return;
      }
    case 's':
      if (strncasecmp(string, "SEND", 4) == 0) {
	mep_send(strip(string, ' '));
	break;
      }
    case 'v':
      if (strncasecmp(string, "VERI", 4) == 0) {
	mep_verify(strip(string, ' '));
	break;
      }
    default:
      sendout("ERRO\r\n");
    }

    free(string);
  }
}

void mepd(void)
{
  pthread_t tid;
  int desc, sock;
  struct sockaddr_in6 addr;
  socklen_t size = sizeof(struct sockaddr_in6);

  logging_on();

  pthread_key_create(&key, NULL);
  pthread_create(&tid, NULL, (void *(*)(void *)) &unixsocket, NULL);

  addr.sin6_family = AF_INET6;
  addr.sin6_port = htons(atoi(PORT));
  addr.sin6_addr = in6addr_any;

  if (gethostname(host, NI_MAXHOST) < 0)
    sprintf(host, "[unknown]");

  if ((sock = socket(PF_INET6, SOCK_STREAM, 0)) < 0) {
    fprintf(stderr, "Unable to get a socket: %s\n",
	    strerror(errno));
    exit(EXIT_FAILURE);
  }

  if (bind(sock, (struct sockaddr *) &addr, size) < 0) {
    fprintf(stderr, "Unable to bind socket: %s\n",
	    strerror(errno));
    exit(EXIT_FAILURE);
  }

  if (listen(sock, MAX_CONNECTIONS) < 0) {
    fprintf(stderr, "Unable to listen on socket: %s\n",
	    strerror(errno));
    exit(EXIT_FAILURE);
  }

  while (1) {
    if ((desc = accept(sock, (struct sockaddr *) &addr, &size))
	< 0) {
      fprintf(stderr, "Unable to accept connections: %s\n",
	      strerror(errno));
      exit(EXIT_FAILURE);
    }

    pthread_create(&tid, NULL, (void *(*)(void *)) &netsocket,
		   (void *) desc);
  }
}

int main(void)
{
  if (fork() == 0)
    mepd();

  return EXIT_SUCCESS;
}
